#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = 'Michael X. Wang'


import os
CURRENT_DIR = os.path.dirname(__file__)


import json
import pickle
import random
import copy
import multiprocessing
from time import time
from math import ceil
from math import floor
from copy import deepcopy

import numpy as np
import pandas as pd
from numpy.random import rand
from Bio import SeqIO
from matplotlib import pyplot as plt
from matplotlib.widgets import Slider
from numpy.random import default_rng

import basic
import design
from ncbi_tools import BLAST_batch_short, ns_simulation
from design import PrimerSetBadnessFast

plt.rcParams.update({'font.family':'Arial'})

PRIMER_DESIGN_LEN = 40
UPPER_GC = 0.75
LOWER_GC = 0.25
LOW_COMPLEXITY = 0.4
CHOICE_RATE = 0.3
RISK_TH = 0.1 # top RISK_TH highest risk primer design regions are considered as loss


def build(fasta_path, var_path, BLAST_db, title, n_cpu):
    '''
    Build the OliVar reference file for tiled sequencing primer design
    Input:
        fasta_path: path to the reference sequence (in fasta format)
        var_path: path to the csv file lising coordinates of SNPs and iSNVs
        BLAST_db: path to the BLAST database
        title: name of the OliVar reference file
        n_cpu: number of threads for multiprocessing
    '''
    # all coordinates are 1-based and inclusive
    word_size = 28
    offset = 14 # word_size should be divisible by offset
    print('building OliVar reference with word_size=%d, offset=%d' % (word_size, offset))

    n_cycle = word_size/offset
    if n_cycle != int(n_cycle):
        raise ValueError('word_size should be divisible by offset.')
    else:
        n_cycle = int(n_cycle)

    for record in SeqIO.parse(fasta_path, 'fasta'):
        seq_raw = str(record.seq)

    # load SNP/iSNV coordinates
    var_arr = np.zeros(len(seq_raw))
    seq_raw = seq_raw.lower() # SNPs in upper case
    if var_path:
        seq_raw = list(seq_raw)
        df = pd.read_csv(var_path, sep=',')
        for i, row in df.iterrows():
            if np.isnan(row['FREQ']):
                freq = 1.0
            else:
                freq = row['FREQ']**0.5
            var_arr[row['START']-1:row['STOP']] += freq
            # capitalize SNPs
            for pos in range(row['START']-1, row['STOP']):
                seq_raw[pos] = seq_raw[pos].upper()
        seq_raw = ''.join(seq_raw)
    else:
        print('no SNPs/iSNVs coordinates provided, skipping...')

    # calculate risk of GC, non-specificity, complexity
    seq_len_temp = (len(seq_raw)//offset) * offset # length of temporary sequence
    start_temp = (len(seq_raw) - seq_len_temp)//2 + 1 # start coordinate of sequence to be processed as words
    stop_temp = start_temp + seq_len_temp - 1

    # fetch all words
    all_word = []
    for pos in range(0, seq_len_temp-word_size+1, offset):
        word_start = start_temp+pos-1
        word = seq_raw[word_start : word_start+word_size]
        all_word.append(word)

    # get GC, complexity, BLAST hits of each word
    with multiprocessing.Pool(processes=n_cpu) as pool:
        print('calculating GC content and sequence complexity...')
        all_gc = np.array(pool.map(basic.get_GC, all_word))
        all_complexity = np.array(pool.map(basic.get_complexity, all_word))
    if BLAST_db:
        all_hits, _ = BLAST_batch_short(all_word, db=BLAST_db, n_cpu=n_cpu, mode='rough') # tabular output
    else:
        print('no BLAST database provided, skipping...')
        all_hits = np.zeros(len(all_word))

    # calculate score array
    start = start_temp + (n_cycle-1)*offset # start coordinate of sequence with score array
    stop = stop_temp - (n_cycle-1)*offset
    seq_len = stop - start + 1 # length of final sequence
    gc_arr = np.zeros(seq_len)
    comp_arr = np.zeros(seq_len)
    hits_arr = np.zeros(seq_len)
    for pos in range(seq_len):
        n = pos//offset
        gc_arr[pos] = np.sum(all_gc[n:n+n_cycle])
        comp_arr[pos] = np.sum(all_complexity[n:n+n_cycle])
        hits_arr[pos] = np.sum(all_hits[n:n+n_cycle])
    gc_arr = gc_arr/n_cycle
    hits_arr = hits_arr/n_cycle

    olv_ref = {
        'seq': seq_raw, 
        'start': start, 
        'stop': stop, 
        'gc_arr': gc_arr, 
        'comp_arr': comp_arr, 
        'var_arr': var_arr[start-1:stop], 
        'hits_arr': hits_arr, 
        'all_hits': all_hits
    }
    with open('%s.olvr' % title, 'wb') as f:
        pickle.dump(olv_ref, f, protocol=4)
        print('reference file saved as %s.olvr' % title)
    return olv_ref


def find_min_loc(risk_arr, start, stop, rng):
    # find the next primer design region within start and stop
    score = []
    loc = np.array(range(start, stop-PRIMER_DESIGN_LEN+2))
    for i in loc:
        score.append(sum(risk_arr[i-1: i-1+PRIMER_DESIGN_LEN]))
    if len(loc) == 1:
        idx_rand = 0
    elif CHOICE_RATE < 1:
        k = int(len(loc)*CHOICE_RATE)
        if k == 0:
            k = 1
        idx_k_lowest = np.argpartition(score, k)[:k]
        idx_rand = rng.choice(idx_k_lowest)
    else:
        idx_rand = rng.choice(range(len(loc)))
    primer_start = loc[idx_rand]
    return primer_start, primer_start+PRIMER_DESIGN_LEN-1, score[idx_rand]


def generate_context(SOI):
    risk_arr, start, stop, max_amp_len, min_amp_len, seed = SOI
    rng = default_rng(seed)
    
    all_context_seq = [] # coordinates of all context sequences
    all_risk = []

    # generate the first pair of primer design region
    fp_start, fp_stop, fp_risk = find_min_loc(risk_arr, start, start+3*PRIMER_DESIGN_LEN-1, rng)
    rp_stop, rp_start, rp_risk = find_min_loc(risk_arr, fp_start+min_amp_len-PRIMER_DESIGN_LEN, fp_start+max_amp_len-1, rng)
    all_context_seq.append((fp_start, fp_stop, rp_stop, rp_start))
    all_risk.append((fp_risk, rp_risk))

    # generate the second pair of primer design region
    left_lim = max(fp_stop+1, rp_start+2*PRIMER_DESIGN_LEN-max_amp_len+1) # leave space for the next fP design region
    fp_start, fp_stop, fp_risk = find_min_loc(risk_arr, left_lim, rp_stop-1, rng)
    left_lim = max(fp_start+min_amp_len-PRIMER_DESIGN_LEN, rp_start+PRIMER_DESIGN_LEN+1)
    right_lim = fp_start+max_amp_len-1
    if right_lim > stop:
        raise ValueError('sequence provided is too short.')
    rp_stop, rp_start, rp_risk = find_min_loc(risk_arr, left_lim, right_lim, rng)
    all_context_seq.append((fp_start, fp_stop, rp_stop, rp_start))
    all_risk.append((fp_risk, rp_risk))

    # design the rest context sequences
    cs_id = 2 # id of the next context sequence to be designed (0-based)
    while True:
        # fp
        # all_context_seq[cs_id-2][3]: rp_start of the last last context sequence
        left_lim = max(all_context_seq[cs_id-2][3]+1, rp_start+2*PRIMER_DESIGN_LEN-max_amp_len+1)
        fp_start, fp_stop, fp_risk = find_min_loc(risk_arr, left_lim, rp_stop-1, rng)
        # rp
        left_lim = max(fp_start+min_amp_len-PRIMER_DESIGN_LEN, rp_start+PRIMER_DESIGN_LEN+1)
        right_lim = fp_start+max_amp_len-1
        if right_lim > stop:
            break
        rp_stop, rp_start, rp_risk = find_min_loc(risk_arr, left_lim, right_lim, rng)

        all_context_seq.append((fp_start, fp_stop, rp_stop, rp_start))
        all_risk.append((fp_risk, rp_risk))
        cs_id += 1

    all_context_seq = np.array(all_context_seq)
    all_risk = np.array(all_risk)
    all_risk_flatten = all_risk.flatten()
    k = int(len(all_risk_flatten) * (1-RISK_TH))
    loss = sum(np.partition(all_risk_flatten, k)[k:]**2)
    #loss = sum(all_risk_flatten[all_risk_flatten>=RISK_TH])
    return all_context_seq, all_risk, loss


def design_context_seq(config):
    ref_path = config['ref_path']
    max_amp_len = config['max_amp_len']
    w_gc = config['w_gc']
    w_complexity = config['w_complexity']
    w_hits = config['w_hits']
    w_var = config['w_var']
    seed = config['seed']
    n_cpu = config['n_cpu']
    
    # set random number generator
    rng_parent = default_rng(seed)
    
    with open(ref_path,  'rb') as f:
        olv_ref = pickle.load(f)

    seq_raw = olv_ref['seq']
    start = olv_ref['start'] # start of design region, 1-based, closed
    stop = olv_ref['stop'] # stop of design region, 1-based, closed
    gc_arr = olv_ref['gc_arr']
    comp_arr = olv_ref['comp_arr']
    var_arr = olv_ref['var_arr']
    hits_arr = olv_ref['hits_arr']
    print('successfully loaded reference file %s' % ref_path)
    
    if config['min_amp_len'] is None:
        min_amp_len = int(max_amp_len*0.9)
    else:
        min_amp_len = config['min_amp_len']
    if min_amp_len < 3*PRIMER_DESIGN_LEN:
        raise ValueError('min_amp_len is too small')

    gc_arr = np.logical_or(gc_arr<LOWER_GC, gc_arr>UPPER_GC).astype(int)
    comp_arr = (comp_arr < LOW_COMPLEXITY).astype(int)
    hits_arr = hits_arr/max(hits_arr) # normalize

    # make arrays the same length as seq_raw
    gc_arr = w_gc*0.5*np.concatenate((np.zeros(start-1), gc_arr, np.zeros(len(seq_raw)-stop)))
    comp_arr = w_complexity*0.5*np.concatenate((np.zeros(start-1), comp_arr, np.zeros(len(seq_raw)-stop)))
    hits_arr = w_hits*np.concatenate((np.zeros(start-1), hits_arr, np.zeros(len(seq_raw)-stop)))
    var_arr = w_var*10*np.concatenate((np.zeros(start-1), var_arr, np.zeros(len(seq_raw)-stop)))

    # construct risk array (first row is risk, second row is coordinate on seq_rawy)
    risk_arr = gc_arr + comp_arr + hits_arr + var_arr

    N = 500*len(risk_arr)//max_amp_len # number of primer sets to generate
    #N = 1
    rand_int = rng_parent.integers(2**32, size=N) # random seeds for each iteration

    # single thread
    # design = [generate_context((risk_arr, start, stop, max_amp_len, min_amp_len, rand_int[i])) for i in tqdm(range(N))]

    # multi threads
    print('reference sequence length: %d' % len(seq_raw))
    print('design region: %d:%d' % (start, stop))
    print('region length: %d' % (stop-start+1))
    print('number of iterations: %d' % N)
    print('designing context sequences...')
    tik = time()
    with multiprocessing.Pool(processes=n_cpu) as pool:
        batch = [(risk_arr, start, stop, max_amp_len, min_amp_len, rand_int[i]) for i in range(N)]
        design = pool.map(generate_context, batch)
    print('finished in %.3fs' % (time()-tik))

    # find the best arangement of primer design regions
    best_design = sorted(design, key=lambda x:x[2])
    all_context_seq, all_risk, loss = best_design[0]
    print('Best total risk: %.3f' % loss)

    # plot
    all_loss = [d[2] for d in design]
    all_loss.sort(reverse=True)
    fig = plt.figure(figsize=(8, 5))
    plt.plot(all_loss)
    plt.title('Optimization of primer design regions', fontsize=14)
    plt.xlabel('iterations (sorted)', fontsize=14)
    plt.ylabel('Total risk', fontsize=14)
    plt.show()

    # plt.hist(all_risk.flatten(), log=True)
    # #plt.xlim(right=32)
    # plt.xlabel('Risk of primer design regions', size=12)
    # plt.ylabel('# primer design regions', size=12)
    # plt.show()
    
    # prepare output
    all_plex_info = {}
    for i, (context_seq, risk) in enumerate(zip(all_context_seq, all_risk)):
        cs = seq_raw[context_seq[0]-1:context_seq[3]] # actual context sequence
        plex_info = {
            'tube': i%2 + 1, 
            'context_seq_coords': (context_seq[0], context_seq[3]), 
            'primers_coords': tuple(context_seq), 
            'insert_coords': (context_seq[1]+1, context_seq[2]-1), 
            'context_seq': cs, 
            'context_seq_revcomp': basic.revcomp(cs), 
            'fP_design': seq_raw[context_seq[0]-1:context_seq[1]], 
            'rP_design': basic.revcomp(seq_raw[context_seq[2]-1:context_seq[3]]), 
            'risk': tuple(risk)
        }
        all_plex_info['amp_%d' % (i+1)] = plex_info
    print('total amplicons: %d' % (i+1))
    cover_start = all_context_seq[0][1]+1
    cover_stop = all_context_seq[-1][2]-1
    print('covered region: %d:%d' % (cover_start, cover_stop))
    print('coverage of reference sequence: %.3f%%' % (100*(cover_stop-cover_start+1)/len(seq_raw)))
    #return seq_raw, risk_arr, all_context_seq, all_risk, all_plex_info
    return all_plex_info, risk_arr, gc_arr, comp_arr, hits_arr, var_arr, all_loss, seq_raw


def get_primer(all_plex_info, config):
    '''
    Generate candidates of fP and rP for each plex
    Input:
        all_plex_info: output of design_context_seq(all_gene_info, config), or load from 'all_plex_info.pkl'
        config: design configurations, load from 'config.json'
    Output:
        all_plex_info and corresponding 'all_plex_info_primer.pkl': add primer candidates to input all_plex_info
    '''
    print('Generating primer candidates...')

    temperature_fP = config['temperature_fP']
    temperature_rP = config['temperature_rP']
    salinity = config['salinity']
    fP_adapter = config['fP_adapter']
    rP_adapter = config['rP_adapter']
    default_fP_setting = config['default_fP_setting']
    default_rP_setting = config['default_rP_setting']
    check_BLAST = False

    # generate primer candidates
    fP_generator = design.primer_generator(temperature_fP, salinity)
    rP_generator = design.primer_generator(temperature_rP, salinity)
    n = 0
    for plex_id, plex_info in all_plex_info.items():
        # set primer prefix
        fP_prefix = fP_adapter
        rP_prefix = rP_adapter
        n += 1
        
        # generate fP
        fP_design = plex_info['fP_design'] # sequence to generate fP candidates
        fP_setting = deepcopy(default_fP_setting)
        fP, fail = fP_generator.get(fP_design, fP_prefix, check_BLAST=check_BLAST, **fP_setting)
        while len(fP) == 0:
            print('Fail to generate fP, update setting: %s' % plex_id)
            # detect failure mode with the most cases
            fail_mode = max(fail, key=fail.get)
            # update setting
            if fail_mode == 'SNP_fail' and fP_setting['check_SNP'] == True:
                fP_setting['check_SNP'] = False
            elif fail_mode == 'GC_fail':
                fP_setting['min_GC'] -= 0.05
                fP_setting['max_GC'] += 0.05
            elif fail_mode == 'complexity_fail':
                fP_setting['min_complexity'] -= 0.05
            # generate primers again
            fP, fail = fP_generator.get(fP_design, fP_prefix, check_BLAST=check_BLAST, **fP_setting)
        plex_info['fP_candidate'] = fP
        plex_info['fP_setting'] = fP_setting
        
        # generate rP
        rP_design = plex_info['rP_design'] # sequence to generate rP candidates
        rP_setting = deepcopy(default_rP_setting)
        rP, fail = rP_generator.get(rP_design, rP_prefix, check_BLAST=check_BLAST, **rP_setting)
        while len(rP) == 0:
            print('Fail to generate rP, update setting: %s' % plex_id)
            # detect failure mode with the most cases
            fail_mode = max(fail, key=fail.get)
            # update setting
            if fail_mode == 'SNP_fail' and rP_setting['check_SNP'] == True:
                rP_setting['check_SNP'] = False
            elif fail_mode == 'GC_fail':
                rP_setting['min_GC'] -= 0.05
                rP_setting['max_GC'] += 0.05
            elif fail_mode == 'complexity_fail':
                rP_setting['min_complexity'] -= 0.05
            # generate primers again
            rP, fail = rP_generator.get(rP_design, rP_prefix, check_BLAST=check_BLAST, **rP_setting)
        plex_info['rP_candidate'] = rP
        plex_info['rP_setting'] = rP_setting
        
    # with open('all_plex_info_primer.pkl', 'wb') as f:
    #     pickle.dump(all_plex_info, f)
    print('Done\n')
    return all_plex_info


def optimize(all_plex_info, config):
    '''
    Optimize fP and rP
    Input:
        all_plex_info: output of get_primer(all_plex_info, config), or load from 'all_plex_info_primer.pkl'
        config: design configurations, load from 'config.json'
    Output:
        all_plex_info and corresponding 'all_plex_info_optimize.pkl': add optimized fP and rP to each plex of input all_plex_info
    '''
    print('Optimizing primer dimers...')
    # set random seed
    np.random.seed(config['seed'])
    random.seed(config['seed'])

    n_tube = 2
    InitSATemp = config['SA_setting']['InitSATemp']
    NUMSTEPS = config['SA_setting']['NUMSTEPS']
    ZEROSTEPS = config['SA_setting']['ZEROSTEPS']
    TimePerStep = config['SA_setting']['TimePerStep']

    # generate primer pair candidates (fP-rP)
    n_pair = []
    for plex_id, plex_info in all_plex_info.items():
        optimize = []
        plex_amp_len = []
        for i, fp in plex_info['fP_candidate'].iterrows():
            for j, rp in plex_info['rP_candidate'].iterrows():
                insert_len = plex_info['insert_coords'][1] - plex_info['insert_coords'][0] + 1
                # check amplicon length
                amp_len = fp['primer_len'] + fp['dist'] + insert_len + rp['dist'] + rp['primer_len']
                plex_amp_len.append(amp_len)
                optimize.append([fp, rp])
        plex_info['optimize'] = optimize
        n_pair.append(len(optimize))
    print('total primer pairs %d' % sum(n_pair))
    print('average pairs per plex %.2f' % (sum(n_pair)/len(n_pair)))
    
    # optimize each tube
    all_lc = []
    tik = time()
    for i_tube in range(1, n_tube+1):
        print('\npool %d, simulated annealing...' % i_tube)
        # plex_id of current tube
        curr_tube = [plex_id for plex_id, plex_info in all_plex_info.items() if plex_info['tube'] == i_tube]

        # load existing primers
        existing_primer = config['existing_primer']

        # randomly select one primer set (one pair each plex)
        curr_index = {}
        curr_fp = {}
        curr_rp = {}
        curr_badness = 0
        for plex_id in curr_tube:
            # randomly select one primer pair
            i = floor(rand() * len(all_plex_info[plex_id]['optimize']))
            curr_index[plex_id] = i
            fp = all_plex_info[plex_id]['optimize'][i][0]
            rp = all_plex_info[plex_id]['optimize'][i][1]
            curr_fp[plex_id] = fp['seq']
            curr_rp[plex_id] = rp['seq']
            curr_badness += fp['badness'] + rp['badness']
        inter_badness, comp_badness = PrimerSetBadnessFast(list(curr_fp.values()), list(curr_rp.values()), existing_primer)
        curr_badness += inter_badness
        print('initial loss = %.3f' % curr_badness)

        # optimization
        learning_curve = []
        SATemp = InitSATemp
        for step in range(NUMSTEPS + ZEROSTEPS):
            print('temperature = %.3f, Loss = %.3f' % (SATemp, curr_badness))
            for t in range(TimePerStep):
                # pick a plex to change primer pair
                mutplex = random.choice(curr_tube) # randomly select a plex_id in current tube
                n_pair = len(all_plex_info[mutplex]['optimize'])
                if n_pair == 1:
                    continue
                mutindex = floor(rand() * n_pair) # index of new primer pair

                # copy current design
                new_index = deepcopy(curr_index)
                new_fp = deepcopy(curr_fp)
                new_rp = deepcopy(curr_rp)
                
                # calculate new badness
                new_index[mutplex] = mutindex
                new_fp[mutplex] = all_plex_info[mutplex]['optimize'][mutindex][0]['seq']
                new_rp[mutplex] = all_plex_info[mutplex]['optimize'][mutindex][1]['seq']
                new_badness = 0
                for plex_id, i in new_index.items():
                    fp = all_plex_info[plex_id]['optimize'][i][0]
                    rp = all_plex_info[plex_id]['optimize'][i][1]
                    new_badness += fp['badness'] + rp['badness']
                inter_badness, new_comp_badness = PrimerSetBadnessFast(list(new_fp.values()), list(new_rp.values()), existing_primer)
                new_badness += inter_badness
                
                # calculate probability of accepting mutation
                if new_badness < curr_badness:
                    accept = True
                else:
                    if SATemp == 0:
                        accept = False
                    elif rand() < 2**((curr_badness - new_badness)/SATemp):
                        accept = True
                    else:
                        accept = False
                        
                # implement acceptance
                if accept:
                    curr_index[mutplex] = new_index[mutplex]
                    curr_fp[mutplex] = new_fp[mutplex]
                    curr_rp[mutplex] = new_rp[mutplex]
                    curr_badness = new_badness
                    comp_badness = new_comp_badness
                    
                # if t % 100 == 0:
                #     print('\ttime = %d, current loss = %.3f' % (t, curr_badness))
                learning_curve.append(curr_badness)
            # reduce SATemp
            SATemp = max(0, (SATemp - InitSATemp/NUMSTEPS))
            
        # save optimization result
        for n, (plex_id, i) in enumerate(curr_index.items()):
            all_plex_info[plex_id]['fP'] = all_plex_info[plex_id]['optimize'][i][0]
            all_plex_info[plex_id]['rP'] = all_plex_info[plex_id]['optimize'][i][1]
            all_plex_info[plex_id]['fP_badness'] = comp_badness[0][n]
            all_plex_info[plex_id]['rP_badness'] = comp_badness[1][n]

        # plot learning curve
        fig = plt.figure(figsize=(8, 5))
        plt.plot(learning_curve)
        plt.title('Primer dimer optimization (pool_%d)' % i_tube, fontsize=14)
        plt.xlabel('step', fontsize=14)
        plt.ylabel('Loss', fontsize=14)
        plt.show()
        all_lc.append(learning_curve)
    
    print('primer dimer optimization finished in %.3fs' % (time()-tik))
    # with open('all_plex_info_optimize.pkl', 'wb') as f:
    #     pickle.dump(all_plex_info, f)
    print('Done\n')
    return all_plex_info, all_lc


def save(all_plex_info, config):
    '''
    save to excel
    Input:
        all_plex_info: output of optimize_2(all_plex_info, config), or load from 'all_plex_info_optimize_2.pkl'
        config: design configurations, load from 'config.json'
    Output:
        'design.xlsx'
    '''
    print('Saving...')
    n_tube = 2

    l_fP_adp = len(config['fP_adapter'])
    l_rP_adp = len(config['rP_adapter'])

    plex_name = [[] for _ in range(n_tube)] # don't use [[]]*n_tube, this will create linked lists
    id_name = [[] for _ in range(n_tube)]

    # sequence specific part of primers
    fp = [[] for _ in range(n_tube)]
    rp = [[] for _ in range(n_tube)]

    # coordinates
    start = [[] for _ in range(n_tube)]
    insert_start = [[] for _ in range(n_tube)]
    insert_end = [[] for _ in range(n_tube)]
    end = [[] for _ in range(n_tube)]

    # sequences
    amp = [[] for _ in range(n_tube)]
    insert = [[] for _ in range(n_tube)]

    # full primer (with adapters and UMI)
    fp_full = [[] for _ in range(n_tube)]
    rp_full = [[] for _ in range(n_tube)]

    # primer interaction badness
    fp_bad = [[] for _ in range(n_tube)]
    rp_bad = [[] for _ in range(n_tube)]

    # BLAST each primer
    fp_blast = [[] for _ in range(n_tube)]
    rp_blast = [[] for _ in range(n_tube)]

    # amplicon BLAST result
    amplicon_blast = [[] for _ in range(n_tube)]

    # assign plexes to tubes and fill out informations
    print('amplicons containing SNPs or iSNVs in primers:')
    for plex_id, plex_info in all_plex_info.items():
        i = int(plex_id.split('_')[-1])
        tube = i % n_tube
        if tube == 0:
            tube = n_tube
        plex_name[tube-1].append(plex_id)
        
        curr_fp = plex_info['fP']
        curr_rp = plex_info['rP']

        curr_fp_seq = curr_fp['seq'][l_fP_adp:]
        curr_rp_seq = curr_rp['seq'][l_rP_adp:]
        if curr_fp_seq != curr_fp_seq.lower() or curr_rp_seq != curr_rp_seq.lower():
            print(plex_id)

        fp[tube-1].append(curr_fp['seq'][l_fP_adp:])
        rp[tube-1].append(curr_rp['seq'][l_rP_adp:])

        pc = plex_info['primers_coords'] # primer design coords ([fp_start, fp_end, rp_start, rp_end])
        curr_start = pc[1] - curr_fp['dist'] - curr_fp['primer_len'] + 1
        curr_insert_start = curr_start + curr_fp['primer_len']
        curr_insert_end = pc[2] + curr_rp['dist'] - 1
        curr_end = curr_insert_end + curr_rp['primer_len']
        start[tube-1].append(curr_start)
        insert_start[tube-1].append(curr_insert_start)
        insert_end[tube-1].append(curr_insert_end)
        end[tube-1].append(curr_end)

        curr_context_seq = plex_info['context_seq']
        amp[tube-1].append(curr_context_seq[curr_start-pc[0]:curr_end-pc[0]+1])
        insert[tube-1].append(curr_context_seq[curr_insert_start-pc[0]:curr_insert_end-pc[0]+1])

        fp_full[tube-1].append(curr_fp['seq'])
        rp_full[tube-1].append(curr_rp['seq'])

        fp_bad[tube-1].append(plex_info['fP_badness'])
        rp_bad[tube-1].append(plex_info['rP_badness'])

        # fp_blast[tube-1].append(curr_fp['BLAST_hits'])
        # rp_blast[tube-1].append(curr_rp['BLAST_hits'])

        # amplicon_blast[tube-1].append(plex_info['BLAST_count'])

    # save to file
    df = []
    for n in range(n_tube):
        df.append(pd.DataFrame({
            'amp_id': plex_name[n], 
            'fP': fp[n], 
            'rP': rp[n], 
            'start': start[n], 
            'insert_start': insert_start[n], 
            'insert_end': insert_end[n], 
            'end': end[n], 
            'amp': amp[n], 
            'insert': insert[n], 
            'fP_full': fp_full[n], 
            'rP_full': rp_full[n], 
            # 'fP_inter_badness': fp_bad[n], 
            # 'rP_inter_badness': rp_bad[n], 
            # 'fP_BLAST': fp_blast[n], 
            # 'rP_BLAST': rp_blast[n], 
            # 'amplicon_BLAST': amplicon_blast[n]
        }))
    with pd.ExcelWriter('%s.xlsx' % config['title']) as writer:
        for n, d in enumerate(df):
            d.to_excel(writer, sheet_name='pool_%d' % (n+1), index=True)
    print('Primers output to %s.xlsx' % config['title'])
    return df


def visualize(design_path):
    with open(design_path,  'rb') as f:
        design_out = pickle.load(f)
    config = design_out['config']
    risk_arr = design_out['risk_arr']
    gc_arr = design_out['gc_arr']
    comp_arr = design_out['comp_arr']
    hits_arr = design_out['hits_arr']
    var_arr = design_out['var_arr']
    df = design_out['df']
    print('Configurations:')
    print(json.dumps(config, indent=4))

    # all offsets are relative to fp_offset
    fp_offset = (0.5*config['w_gc'] + 0.5*config['w_complexity'] + config['w_hits'])/6
    fp_rp_diff = 0.1

    # draw arrow head
    fp_head = (np.array([7, 6, 5, 4, 3, 2, 1, 0]) + 90)/90
    rp_head = (np.array([0, -1, -2, -3, -4, -5, -6, -7]) + 90)/90
    fp_head = fp_offset * fp_head
    rp_head = fp_offset * rp_head - fp_rp_diff*fp_offset
    head_len = len(fp_head)

    # to hide extra scatter below x axis
    fp_arr_1 = np.zeros(len(risk_arr)) - 1 + fp_offset
    rp_arr_1 = np.zeros(len(risk_arr)) - 1 + fp_offset
    fp_arr_2 = np.zeros(len(risk_arr)) - 1
    rp_arr_2 = np.zeros(len(risk_arr)) - 1

    fp_head_arr_1 = np.zeros(len(risk_arr)) - 1 + fp_offset
    rp_head_arr_1 = np.zeros(len(risk_arr)) - 1 + fp_offset
    fp_head_arr_2 = np.zeros(len(risk_arr)) - 1
    rp_head_arr_2 = np.zeros(len(risk_arr)) - 1

    rp_offset = (1-fp_rp_diff) * fp_offset + 1
    fp_offset += 1
    fp_head += 1
    rp_head += 1

    # calcualte arrays to scatter
    for i, row in df[0].iterrows():
        fp_arr_1[row['start']-1: row['insert_start']-1] += fp_offset
        rp_arr_1[row['insert_end']: row['end']] += rp_offset
        fp_head_arr_1[row['insert_start']-1-head_len: row['insert_start']-1] += fp_head
        rp_head_arr_1[row['insert_end']: row['insert_end']+head_len] += rp_head
    for i, row in df[1].iterrows():
        fp_arr_2[row['start']-1: row['insert_start']-1] += fp_offset
        rp_arr_2[row['insert_end']: row['end']] += rp_offset
        fp_head_arr_2[row['insert_start']-1-head_len: row['insert_start']-1] += fp_head
        rp_head_arr_2[row['insert_end']: row['insert_end']+head_len] += rp_head

    r = np.arange(len(risk_arr))

    # Setting Plot and Axis variables as subplots()
    # function returns tuple(fig, ax)
    Plot, Axis = plt.subplots()
    #Plot.set_figheight(5)
    Plot.set_figheight(10)
    Plot.set_figwidth(20)

    # Adjust the bottom size according to the
    # requirement of the user
    plt.subplots_adjust(bottom=0.25)

    # plot the x and y using plot function
    layer0 = gc_arr
    layer1 = layer0 + comp_arr
    layer2 = layer1 + hits_arr
    layer3 = layer2 + var_arr
    l1 = plt.fill_between(r, layer0, alpha=0.5, color='C0', edgecolor=None, label='extreme GC')
    l2 = plt.fill_between(r, layer1, layer0, alpha=0.5, color='C1', edgecolor=None, label='low complexity')
    l3 = plt.fill_between(r, layer2, layer1, alpha=0.5, color='C2', edgecolor=None, label='non-specificity')
    l4 = plt.fill_between(r, layer3, layer2, alpha=0.5, color='C3', edgecolor=None, label='SNPs')
    l5, = plt.plot([0, 1], [-1, -1], 'k', linewidth=2, label='pool 1 (upper)')
    l6, = plt.plot([0, 1], [-1, -1], 'k', linewidth=2, label='pool 2 (lower)')
    plt.legend(handles=[l1, l2, l3, l4, l5, l6], loc='upper right', fontsize=16)

    plt.scatter(r, fp_arr_1, 2, 'k', marker='.')
    plt.scatter(r, rp_arr_1, 2, 'k', marker='.')
    plt.scatter(r, fp_head_arr_1, 2, 'k', marker='.')
    plt.scatter(r, rp_head_arr_1, 2, 'k', marker='.')
    plt.scatter(r, fp_arr_2, 2, 'k', marker='.')
    plt.scatter(r, rp_arr_2, 2, 'k', marker='.')
    plt.scatter(r, fp_head_arr_2, 2, 'k', marker='.')
    plt.scatter(r, rp_head_arr_2, 2, 'k', marker='.')

    plt.title('risk components are stacked together', loc='right', size=16)
    #plt.title('Sars-Cov-2 (S gene)', size=20)
    plt.ylim(bottom=0, top=(fp_offset-1)*6)
    plt.xlabel('position', size=20)
    plt.xticks(fontsize=14)
    plt.ylabel('risk', size=20)
    plt.yticks(fontsize=14)

    # Choose the Slider color
    slider_color = 'White'

    # Set the axis and slider position in the plot
    axis_position = plt.axes([0.2, 0.1, 0.65, 0.03],
                            facecolor = slider_color)
    slider_position = Slider(axis_position,
                            '', 0.1, len(risk_arr))

    # update() function to change the graph when the
    # slider is in use
    def update(val):
        pos = slider_position.val
        Axis.axis([pos, pos+2000, 0, (fp_offset-1)*6])
        #Axis.axis([pos, pos+2000, 0, 0.77])
        Plot.canvas.draw_idle()

    # update function called using on_changed() function
    slider_position.on_changed(update)

    # Display the plot
    plt.show()


def PCR_tiling(config):
    all_plex_info, risk_arr, gc_arr, comp_arr, hits_arr, var_arr, all_loss, seq_raw = design_context_seq(config)
    all_plex_info_primer = get_primer(all_plex_info, config)
    all_plex_info_optimize, learning_curve = optimize(all_plex_info_primer, config)
    df = save(all_plex_info_optimize, config)

    # save design file
    design_out = {
        'config': config, 
        'risk_arr': risk_arr, 
        'gc_arr': gc_arr, 
        'comp_arr': comp_arr, 
        'hits_arr': hits_arr, 
        'var_arr': var_arr, 
        'all_loss': all_loss, 
        'seq': seq_raw, 
        'df': df, 
        'all_plex_info': all_plex_info, 
        'learning_curve': learning_curve
    }
    design_path = '%s.olvd' % config['title']
    with open(design_path, 'wb') as f:
        pickle.dump(design_out, f, protocol=4)
        print('design file saved as %s.olvd' % config['title'])
    #visualize(design_path)


def validate(design_path, pool, BLAST_db, max_amp_len=1500, temperature=60, n_cpu=1):
    seq_list = []
    seq_names = []
    df = pd.read_excel(design_path, pool, index_col=None)
    for i, row in df.iterrows():
        seq_list.extend([row['fP'], row['rP']])
        seq_names.extend(['%s_fP' % row['amp_id'], '%s_rP' % row['amp_id']])

    # non-specific simulation
    df_ns, df_count, all_hits = ns_simulation(BLAST_db, seq_list, seq_names, max_amp_len, n_cpu)
    df_ns.to_csv('%s_non-specific_amp.csv' % pool)
    df_count.to_csv('%s_non-specific_pair.csv' % pool)

    # calculate Badness
    #all_rep = [seq2arr(p) for p in seq_list]
    _, all_bad = PrimerSetBadnessFast(seq_list)
    all_bad = all_bad[0]

    # calculate dG
    gen = design.primer_generator(temperature=temperature, salinity=0.18)
    all_dG = [gen.dG_init+gen.StacksDG(p) for p in seq_list]

    df_val = pd.DataFrame({
        'name': seq_names, 
        'seq': seq_list, 
        'length': [len(p) for p in seq_list], 
        '%GC': [basic.get_GC(p) for p in seq_list], 
        'complexity': [basic.get_complexity(p) for p in seq_list], 
        'dG (%dC)' % temperature: all_dG, 
        'self_align': [design.WAlignScore(p) for p in seq_list], 
        'dimer_score': all_bad, 
        'BLAST_hits': all_hits
    })
    df_val.to_csv('%s.csv' % pool)
    return df_val, df_ns, df_count
